import numpy as np

from ..tests import typecheck_series
from ..tests import typecheck_vector
from .utils import angle_of
from .utils import get_2D_rotation_matrix


def circle_to_double_shell(vector):
    try:
        typecheck_vector(vector)
    except AssertionError:
        typecheck_series(vector)
    angle = angle_of(vector)
    # Scalar multiple needs to be some pi-periodic function of angle
    sawtooth = lambda theta: (theta) % np.pi
    scalar_multiple = sawtooth(angle)
    return scalar_multiple * vector


def circle_to_diamond(vector):
    try:
        typecheck_vector(vector)
    except AssertionError:
        typecheck_series(vector)
    angle = angle_of(vector)
    scalar_multiple = 1 / (np.abs(np.cos(angle)) + np.abs(np.sin(angle)))
    return scalar_multiple * vector


unit_diagonal = np.array([np.sqrt(2) / 2, np.sqrt(2) / 2])
projection_onto_diagonal = lambda vector: np.dot(vector, unit_diagonal) * unit_diagonal

reflection_matrix = np.array([[-1, 0], [0, 1]])
reflect_over_vertical_axis = lambda vector: reflection_matrix @ vector

rotate_45_ccw = lambda vector: get_2D_rotation_matrix(45) @ vector

shear_matrix = np.array([[1, 0], [3, 1]])
shear_vertically = lambda vector: shear_matrix @ vector



import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from ipywidgets import fixed

from ..plotting import does_function_preserve_addition
from ..plotting import does_function_preserve_scalar_multiplication


class scalar_multiplication_widget(widgets.VBox):
    def __init__(self, callback):

        # Set up the slider for the scalar
        scalar_values = np.round(np.linspace(0, 5, num=101), decimals=2)

        scalar_slider_options = {
            "options": scalar_values,
            "value": 1,
            "description": "scalar",
            "continuous_update": False,
        }
        self.scalar_slider = widgets.SelectionSlider(**scalar_slider_options)

        # Set up the slider for the vector
        angle_values = np.round(np.linspace(0, 2 * np.pi, num=100), decimals=2)

        vector_slider_options = {
            "options": angle_values,
            "value": 0,
            "description": "vector (angle from horizontal axis in radians)",
            "continuous_update": False,
        }
        self.vector_slider = widgets.SelectionSlider(**vector_slider_options)

        # Put both sliders on top
        self.sliders = widgets.VBox(children=[self.scalar_slider, self.vector_slider])

        # widget that actually contains the resulting plot
        self.plot_widget = widgets.interactive_output(
            does_function_preserve_scalar_multiplication,
            {
                "angle": self.vector_slider,
                "scalar": self.scalar_slider,
                "function": fixed(callback),
            },
        )

        super().__init__(children=[self.sliders, self.plot_widget])


class addition_widget(widgets.VBox):
    def __init__(self, callback):

        # Set up the sliders for both vectors
        angle_values = np.round(np.linspace(0, 2 * np.pi, num=101), decimals=2)
        vector1_slider_options = {
            "options": angle_values,
            "value": 0,
            "description": "vector 1",
            "continuous_update": False,
        }
        vector2_slider_options = {
            "options": angle_values,
            "value": np.round(np.pi / 2, decimals=2),
            "description": "vector 2",
            "continuous_update": False,
        }
        self.vector1_slider = widgets.SelectionSlider(**vector1_slider_options)
        self.vector2_slider = widgets.SelectionSlider(**vector2_slider_options)

        # Put both sliders on top
        self.sliders = widgets.VBox(children=[self.vector1_slider, self.vector2_slider])

        # widget that actually contains the resulting plot
        self.plot_widget = widgets.interactive_output(
            does_function_preserve_addition,
            {
                "angle1": self.vector1_slider,
                "angle2": self.vector2_slider,
                "function": fixed(callback),
            },
        )

        super().__init__(children=[self.sliders, self.plot_widget])



import numpy as np

from .tests import typecheck_column_vector
from .tests import typecheck_row_vector
from .tests import typecheck_vector


def vec_to_col(vector):
    typecheck_vector(vector)
    return vector.reshape((-1, 1))


def vec_to_row(vector):
    typecheck_vector(vector)
    return vector.reshape((1, -1))


def col_to_vec(column):
    typecheck_column_vector(column)
    return column.reshape((-1))


def row_to_vec(row):
    typecheck_row_vector(row)
    return row.reshape((-1))


from .tests import typecheck_matrix
from .tests import typecheck_array


def ignore_floating_point_error(array):
    typecheck_array(array)
    epsilon = 10 * np.finfo(np.float).eps
    # This line might be in-place and have unintended side effects
    array[np.abs(array) < epsilon] = 0
    return array


def permute_rows(matrix, permutation):
    """
    Given a permutation of the index set corresponding to the number of rows of a given matrix,
    permute the rows of the matrix using that permutation.
    """
    typecheck_matrix(matrix)
    try:
        typecheck_vector(permutation)
    except AssertionError:
        assert (
            type(permutation) == list
        ), "Permutation must be represented by a one-modal NumPy array or a list"
    # Permutation is of the same index set as number of rows of matrix
    height = matrix.shape[0]
    assert len(permutation) == height and np.allclose(
        np.sort(permutation), np.arange(height)
    ), "Permutation must be of the indices of the rows."
    return matrix[np.argsort(permutation), :]


def permute_columns(matrix, permutation):
    """
    Given a permutation of the index set corresponding to the number of columns of a given matrix,
    permute the columns of the matrix using that permutation.
    """
    typecheck_matrix(matrix)
    try:
        typecheck_vector(permutation)
    except AssertionError:
        assert (
            type(permutation) == list
        ), "Permutation must be represented by a one-modal NumPy array or a list"
    # Permutation is of the same index set as number of columns of matrix
    width = matrix.shape[1]
    assert len(permutation) == width and np.allclose(
        np.sort(permutation), np.arange(width)
    ), "Permutation must be of the indices of the columns."
    return matrix[:, np.argsort(permutation)]


# Helper function to stringify matrices in a prettier way
def stringify_matrix(A):
    typecheck_matrix(A)
    return "\n".join(
        [
            " ".join(["{:7.4f}".format(A[i, j]) for j in range(A.shape[1])])
            for i in range(A.shape[0])
        ]
    )



from .datasets import circuit_courts
from .datasets import discretized_functions
from .datasets import eastern_cities
from .datasets import federal_reserve_districts
from .datasets import iris
from .datasets import metropolises
from .datasets import mountains
from .datasets import states
from .datasets import students
from .datasets import western_cities



import matplotlib.pyplot as plt
import numpy as np


def scree_plot_factory(sorted_eigenvalues, proportion_explained, ylim=[-0.1, 1.1]):
    # Calculate the vector containing the proportion of
    # total variance explained by each principal component.
    out = proportion_explained(sorted_eigenvalues)

    number_of_pcs = len(sorted_eigenvalues)
    xlim = [-0.1, number_of_pcs + 0.1]

    # Plot the scree plot based on this information.
    fig = plt.figure()
    axes = plt.gca()
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    fig.suptitle("Proportion of variance explained by each principal component")
    plt.xlabel("$k$th principal component")
    plt.ylabel("Proportion of variance")

    xticks = np.arange(np.ceil(xlim[0]), np.floor(xlim[1]))
    plt.xticks(xticks)

    plt.plot(out, "-o")
    plt.show()


def cumulative_variance_plot_factory(
    sorted_eigenvalues, proportion_explained, cumulative_explained
):
    proportions = proportion_explained(sorted_eigenvalues)
    cumulative_proportions = cumulative_explained(sorted_eigenvalues)

    number_of_pcs = len(sorted_eigenvalues)

    with plt.style.context("seaborn-whitegrid"):
        plt.figure(figsize=(6, 4))

        plt.bar(
            range(number_of_pcs),
            proportions,
            alpha=0.5,
            width=1,
            align="edge",
            label="Individual explained variance",
        )
        plt.step(
            range(number_of_pcs),
            cumulative_proportions,
            where="post",
            label="Cumulative explained variance",
        )
        plt.ylabel("Proportion of explained variance")
        plt.xlabel("Principal components ranked by eigenvalues")
        xticks = np.arange(0, number_of_pcs)
        plt.xticks(xticks)
        plt.title("Variance explained by the PCs")
        plt.legend(loc="best")
        plt.tight_layout()



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from ..tests import typecheck_matrix
from ..tests import typecheck_vector


def plot_3D_vector(ax, vector, label="", color="k"):
    typecheck_vector(vector)
    assert len(vector) == 3
    origin = np.array([0, 0, 0])
    ax.quiver(*origin, *vector, color=color)
    if label:
        ax.text(*vector, label, color=color)


def plot_parallelepiped(ax, A):
    P = np.array(
        [
            [-1, -1, -1],
            [1, -1, -1],
            [1, 1, -1],
            [-1, 1, -1],
            [-1, -1, 1],
            [1, -1, 1],
            [1, 1, 1],
            [-1, 1, 1],
        ]
    )
    P = (A @ P.T).T
    verts = [
        [P[0], P[1], P[2], P[3]],
        [P[4], P[5], P[6], P[7]],
        [P[0], P[1], P[5], P[4]],
        [P[2], P[3], P[7], P[6]],
        [P[1], P[2], P[6], P[5]],
        [P[4], P[7], P[3], P[0]],
    ]
    polygon_options = {"alpha": 0.1, "linewidths": 1, "edgecolors": "r"}
    pc = Poly3DCollection(verts, **polygon_options)
    pc.set_facecolor("cyan")
    ax.add_collection3d(pc)
    ax.set_xlabel("$x_1$")
    ax.set_ylabel("$x_2$")
    ax.set_zlabel("$x_3$")


def plot_3D_spectral_demo(
    transformation_matrix, eigenbasis_matrix, plot_size=3, eigvec_color="m"
):
    typecheck_matrix(transformation_matrix)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set_xlim(-plot_size, plot_size)
    ax.set_ylim(-plot_size, plot_size)
    ax.set_zlim(-plot_size, plot_size)

    I = np.eye(3)
    transformed_eigenbasis_matrix = transformation_matrix @ eigenbasis_matrix

    plot_parallelepiped(ax, transformation_matrix)

    # plot unit vectors under transformation
    plot_3D_vector(ax, transformation_matrix[:, 0], "e1")
    plot_3D_vector(ax, transformation_matrix[:, 1], "e2")
    plot_3D_vector(ax, transformation_matrix[:, 2], "e3")

    # plot eigenbasis under transformation
    plot_3D_vector(ax, transformed_eigenbasis_matrix[:, 0], "v1", color=eigvec_color)
    plot_3D_vector(ax, transformed_eigenbasis_matrix[:, 1], "v2", color=eigvec_color)
    plot_3D_vector(ax, transformed_eigenbasis_matrix[:, 2], "v3", color=eigvec_color)



import numpy as np
import pandas as pd


def sort_eigs(eigvals, eigvecs):
    idx = eigvals.argsort()[::-1]
    eigvals = eigvals[idx]
    eigvecs = eigvecs[:, idx]
    return eigvals, eigvecs


def ball_to_box_ratio(num_trials, callback):
    num_hits = 0
    for _ in range(num_trials):
        dart = callback()
        if dart["in_ball"]:
            num_hits += 1
    return num_hits / num_trials


def throw_darts(num_trials, callback):
    darts = np.array([])

    for _ in range(num_trials):
        dart = callback()

        d = len(dart["coordinates"])

        if d == 1:
            dart["coordinates"] = np.pad(dart["coordinates"], (0, 2))
        elif d == 2:
            dart["coordinates"] = np.pad(dart["coordinates"], (0, 1))
        elif d == 3:
            pass
        else:
            raise ValueError("Dart must be two or three dimensional")

        dart = np.append(dart["coordinates"], dart["in_ball"])
        darts = np.append(darts, dart)

    darts = darts.reshape([-1, 4])
    darts = pd.DataFrame(darts).rename(
        columns={
            0: "First Dimension",
            1: "Second Dimension",
            2: "Third Dimension",
            3: "Inside Ball",
        }
    )
    return darts



import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from ipywidgets import fixed
from ipywidgets import interactive
from ipywidgets import SelectionSlider

from ..plotting import p_ball_plot
from ..utils import throw_darts


def p_ball_widget(callback, concave=False):
    """
    `callback` is intended to be the dart_did_land_in_ball function students making in the notebook
    `concave` dictates whether slider is for non-norm values, 0 < p < 1
    """
    figure = plt.figure()
    axis = figure.add_subplot(111, aspect="equal", xlim=[-1.1, 1.1], ylim=[-1.1, 1.1])

    if concave:
        values_of_p = np.round(np.linspace(0, 1, 101)[1:], decimals=2)
        start_value = 0.5
    else:
        non_log_scale = np.concatenate(
            [
                np.linspace(0, 1, 51),
                np.linspace(1, 3, 26)[1:],
                np.linspace(3, 10, 26)[1:],
            ]
        )
        values_of_p = np.round(2 ** non_log_scale, decimals=2)
        start_value = 2

    slider_options = {
        "options": values_of_p,
        "value": start_value,
        "description": "p",
        "continuous_update": False,
    }
    plot_options = {
        "figure": fixed(figure),
        "axis": fixed(axis),
        "trials": fixed(10000),
        "callback": fixed(callback),
    }

    widget = interactive(
        p_ball_plot, p=SelectionSlider(**slider_options), **plot_options
    )
    return widget


class p_ball_widget_3d(widgets.VBox):
    def __init__(self, callback, num_trials=10000):
        self.callback = callback
        self.num_trials = num_trials

        non_log_scale = np.concatenate(
            [
                np.linspace(0, 1, 51),
                np.linspace(1, 3, 26)[1:],
                np.linspace(3, 10, 26)[1:],
            ]
        )
        values_of_p = np.round(2 ** non_log_scale, decimals=2)
        slider_options = {
            "options": values_of_p,
            "value": 2,
            "description": "p",
            "continuous_update": False,
        }
        self.slider = widgets.SelectionSlider(**slider_options)
        self.slider.observe(self.p_event, names="value")

        self.generate_data(self.slider.value)

        axis_settings = {"range": [-1.0, 1.0], "tickmode": "linear", "dtick": 0.5}
        axis_titles = {
            "xaxis_title": "$x_1$",
            "yaxis_title": "$x_2$",
            "zaxis_title": "$x_3$",
        }

        figure = go.Figure()
        figure.add_trace(
            go.Scatter3d(
                x=self.darts_in["First Dimension"].values,
                y=self.darts_in["Second Dimension"].values,
                z=self.darts_in["Third Dimension"].values,
                mode="markers",
                marker={"size": 1.5, "opacity": 0.7},
            )
        )
        figure.add_trace(
            go.Scatter3d(
                x=self.darts_out["First Dimension"].values,
                y=self.darts_out["Second Dimension"].values,
                z=self.darts_out["Third Dimension"].values,
                mode="markers",
                marker={"size": 1.5, "opacity": 0.7},
            )
        )
        figure.update_layout(
            title="The ratio of the L-{} unit ball to box is {}".format(
                self.slider.value, self.ball_to_box_ratio
            ),
            showlegend=False,
            # https://plot.ly/python/custom-buttons/
            updatemenus=[
                go.layout.Updatemenu(
                    type="buttons",
                    direction="right",
                    buttons=[
                        {
                            "label": "All Points",
                            "method": "update",
                            "args": [
                                {"visible": [True, True]},
                                # {"title" : "All Points (Including Those Outside the Ball)"}
                            ],
                        },
                        {
                            "label": "Inside Only",
                            "method": "update",
                            "args": [
                                {"visible": [True, False]},
                                # {"title": "Points Inside the Ball Only"}
                            ],
                        },
                    ],
                )
            ],
        )
        figure.update_layout(
            scene={
                "aspectmode": "cube",
                "xaxis": axis_settings,
                "yaxis": axis_settings,
                "zaxis": axis_settings,
                "xaxis_title": "x1",
                "yaxis_title": "x2",
                "zaxis_title": "x3",
            }
        )
        self.figure = go.FigureWidget(figure)

        # 21 chunks, each of length 5% except for two edge chunks of length 2.5%
        slice_slider_options = {
            "value": 0.5,
            "step": 0.05,
            "min": 0,
            "max": 1.0,
            "continuous_update": False,
        }
        self.slice_slider = widgets.FloatSlider(**slice_slider_options)
        self.slice_slider.observe(self.slice_event, names="value")

        self.slice_data(self.slice_slider.value)
        slice_figure = go.Figure(
            data=go.Scatter3d(
                x=self.darts_slice["First Dimension"].values,
                y=self.darts_slice["Second Dimension"].values,
                z=self.darts_slice["Third Dimension"].values,
                mode="markers",
                marker={"size": 1.5, "opacity": 0.7},
            )
        )
        slice_figure.update_layout(
            title="Cross section centered at {}% plus or minus 2.5%".format(
                int(round(self.slice_slider.value * 100))
            )
        )

        slice_figure.update_layout(
            scene={
                "aspectmode": "cube",
                "xaxis": axis_settings,
                "yaxis": axis_settings,
                "zaxis": axis_settings,
                "xaxis_title": "x1",
                "yaxis_title": "x2",
                "zaxis_title": "x3",
            }
        )

        self.slice_figure = go.FigureWidget(slice_figure)

        super().__init__(
            children=[self.slider, self.figure, self.slice_slider, self.slice_figure]
        )

    def p_event(self, change):
        self.update_p(change.new)
        self.update_slice(self.slice_slider.value)

    def slice_event(self, change):
        self.update_slice(change.new)

    def update_p(self, p):
        # https://plot.ly/python/slider-widget/#sliders-with-3d-plots
        self.generate_data(p)

        self.figure.data[0].x = self.darts_in["First Dimension"].values
        self.figure.data[0].y = self.darts_in["Second Dimension"].values
        self.figure.data[0].z = self.darts_in["Third Dimension"].values

        self.figure.data[1].x = self.darts_out["First Dimension"].values
        self.figure.data[1].y = self.darts_out["Second Dimension"].values
        self.figure.data[1].z = self.darts_out["Third Dimension"].values

        self.figure.layout.title.text = "The ratio of the L-{} unit ball to box is {}".format(
            p, self.ball_to_box_ratio
        )

    def update_slice(self, quantile_number):
        self.slice_data(quantile_number)

        self.slice_figure.data[0].x = self.darts_slice["First Dimension"].values
        self.slice_figure.data[0].y = self.darts_slice["Second Dimension"].values
        self.slice_figure.data[0].z = self.darts_slice["Third Dimension"].values

        self.slice_figure.layout.title.text = "Cross section centered at {}% plus or minus 2.5%".format(
            int(round(quantile_number * 100))
        )

    def generate_data(self, p):
        self.darts = throw_darts(self.num_trials, lambda: self.callback(p))
        self.darts_in = self.darts.loc[self.darts["Inside Ball"] == True]
        self.darts_out = self.darts.loc[self.darts["Inside Ball"] == False]

        self.ball_to_box_ratio = len(self.darts_in) / len(self.darts)

    def slice_data(self, quantile_number):
        # 21 chunks, each of length 5% except for two edge chunks of length 2.5%
        # then take data that's within +- 2.5% of the center value
        upper_bound = 2 * (quantile_number + 0.025) - 1
        lower_bound = 2 * (quantile_number - 0.025) - 1
        # slice along xz plane since z plane is compressed out for 2d plots
        # and want slicing for 2d to also work and return 1d things
        upper_mask = self.darts_in["Second Dimension"] <= upper_bound
        lower_mask = self.darts_in["Second Dimension"] >= lower_bound
        self.darts_slice = self.darts_in.loc[lower_mask & upper_mask]
        self.darts_slice["Second Dimension"].values[:] = 2 * quantile_number - 1



import numpy as np
import pandas as pd
import plotly.graph_objects as go

from ..tests import typecheck_vector


def scatter_3D(data):
    fig = go.Figure(
        data=[
            go.Scatter3d(
                x=data.iloc[:, 0].values,
                y=data.iloc[:, 1].values,
                z=data.iloc[:, 2].values,
                mode="markers",
                marker=dict(
                    size=2,
                    color=data.iloc[:, 3].values,
                    colorscale="Rainbow",  # choose a colorscale
                    opacity=0.8,
                ),
            )
        ]
    )
    fig.update_layout(
        scene=dict(), width=700, height=700, margin=dict(r=10, l=10, b=10, t=10)
    )
    fig.show()


def scatter_3D_collinear(matrix, xlim=1, num=10000):
    darts = np.array([])
    for i in range(num):
        dart = np.random.rand(2) * xlim - xlim / 2
        if dart[0] > 0 and dart[1] > 0:
            color = 0
        elif dart[0] > 0 and dart[1] < 0:
            color = 1
        elif dart[0] < 0 and dart[1] > 0:
            color = 2
        else:
            color = 3
        darts = np.append(darts, np.append(dart, np.array([0, color])))
    darts = darts.reshape(-1, 4)
    darts[:, :2] = darts[:, :2] @ matrix.T
    darts = pd.DataFrame(darts)
    scatter_3D(darts)
    return  # darts


def plot_vectors_in_space(*vectors):
    for vector in vectors:
        typecheck_vector(vector)

    column_names = {0: "First Dimension", 1: "Second Dimension", 2: "Third Dimension"}

    fig = go.Figure()

    for vector in vectors:
        raw_data = np.append(vector, np.array([0, 0, 0]))
        data = pd.DataFrame(raw_data.reshape(-1, 3)).rename(columns=column_names)
        fig.add_trace(
            go.Scatter3d(
                x=data["First Dimension"].values,
                y=data["Second Dimension"].values,
                z=data["Third Dimension"].values,
                mode="lines",
                line=dict(width=2),
            )
        )
    fig.update_layout(
        scene=dict(
            xaxis=dict(nticks=4, range=[-1, 1],),
            yaxis=dict(nticks=4, range=[-1, 1],),
            zaxis=dict(nticks=4, range=[-1, 1],),
        ),
        width=700,
        margin=dict(r=20, l=10, b=10, t=10),
    )
    fig.update_layout(showlegend=False)
    fig.show()


from ..utils import throw_darts


def p_ball_plot_3d(num_trials, callback, inside_only=False):
    """
    `callback` is supposed to be the `throw_dart` function
    which students write, possibly a lambda version with
    some parameters filled in
    """
    darts = throw_darts(num_trials, callback)

    if inside_only == True:
        darts = darts.loc[darts["Inside Ball"] == True]

    fig = go.Figure(
        data=[
            go.Scatter3d(
                x=darts["First Dimension"].values,
                y=darts["Second Dimension"].values,
                z=darts["Third Dimension"].values,
                mode="markers",
                marker=dict(
                    size=1,
                    color=darts["Inside Ball"].values,
                    colorscale="Viridis",  # choose a colorscale
                    opacity=0.8,
                ),
            )
        ]
    )
    fig.show()



from ..utils import throw_darts


def p_ball_plot(p, figure, axis, callback, trials=10000):
    """
    `callback` is meant to be the `throw_dart`
    function students write by themselves
    """
    # Need this line for using with widget
    axis.clear()

    axis.set_xlim(-1.1, 1.1)
    axis.set_ylim(-1.1, 1.1)
    axis.set_xlabel("$x_1$")
    axis.set_ylabel("$x_2$")
    axis.set_title("Scatter plot of points in the $p$-ball")
    axis.set_aspect("equal")

    darts = throw_darts(trials, lambda: callback(p))
    darts = darts.loc[darts["Inside Ball"] == True]
    x1s = darts["First Dimension"].values
    x2s = darts["Second Dimension"].values
    axis.scatter(x1s, x2s)
    figure.canvas.draw()



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

##### TODO: Implement something like this in plotly and compare whether it looks better?


class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


def plot_3D_with_arrows(data, arrows=None, size=5, title="Scatter Plot of Data"):
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection="3d")
    ax.set_xlim(-size, size)
    ax.set_ylim(-size, size)
    ax.set_zlim(-size, size)
    plt.rcParams["legend.fontsize"] = 10
    plot_options = {
        "markersize": 7,
        "color": "blue",
        "alpha": 0.2,
        "label": "observation",
    }
    ax.plot(data[0, :], data[1, :], data[2, :], "o", **plot_options)
    ax.set_title(title)
    ax.legend(loc="upper right")

    ax.set_xlabel("$x_1$")
    ax.set_ylabel("$x_2$")
    ax.set_zlabel("$x_3$")

    if arrows is not None:
        mu_x = np.mean(data[0, :])
        mu_y = np.mean(data[1, :])
        mu_z = np.mean(data[2, :])

        arrow_options = {
            "mutation_scale": 25,
            "lw": 3,
            "arrowstyle": "->",
            "color": "r",
        }
        for arrow in arrows:
            ax.add_artist(
                Arrow3D(
                    [mu_x, arrow[0]],
                    [mu_y, arrow[1]],
                    [mu_z, arrow[2]],
                    **arrow_options
                )
            )

    plt.show()



from .demo import circle_to_diamond
from .demo import circle_to_double_shell
from .demo import projection_onto_diagonal
from .demo import reflect_over_vertical_axis
from .demo import rotate_45_ccw
from .demo import shear_vertically
from .utils import apply_func_to_basis
from .utils import get_2D_rotation_matrix
from .utils import get_parts_for_diagonalization_demo
from .utils import Rx
from .utils import Ry
from .utils import Rz



import matplotlib.pyplot as plt
import numpy as np

from ..arrays import col_to_vec
from ..arrays import row_to_vec
from ..tests import typecheck_vector


def abline(slope, intercept):
    """Plot a line from slope and intercept"""
    axes = plt.gca()
    x_vals = np.array(axes.get_xlim())
    y_vals = intercept + slope * x_vals
    plt.plot(x_vals, y_vals, "--")


def timeseries_plot(discretized_function_vector, title="Time Series Plot", ylim=None):
    try:
        typecheck_vector(discretized_function_vector)
    except AssertionError:
        try:
            col_to_vec(discretized_function_vector)
        except AssertionError:
            row_to_vec(discretized_function_vector)

    xvalues = np.arange(len(discretized_function_vector))
    plt.plot(xvalues, discretized_function_vector)
    plt.title(title)
    plt.xlabel("Coordinate number")
    plt.ylabel("Value at coordinate")
    plt.xlim([-1, len(discretized_function_vector) + 1])
    if ylim:
        plt.ylim(ylim)
    else:
        plt.ylim(
            [
                np.min(discretized_function_vector) - 0.1,
                np.max(discretized_function_vector) + 0.1,
            ]
        )



import matplotlib.pyplot as plt
import numpy as np


def pairs_histogram_factory(n, m, name, generating_function):
    assert (
        name == "angle" or name == "distance"
    ), "This is implemented only for angles or distances between pairs of points."
    geometrical_quantity = generating_function(n, m)
    print(
        "The mean of the {}s in R^{} is {}".format(
            name, n, np.mean(geometrical_quantity)
        )
    )
    print(
        "The stdev of the {}s in R^{} is {}".format(
            name, n, np.std(geometrical_quantity)
        )
    )
    plt.hist(geometrical_quantity)
    if name == "angle":
        plt.xlim(0, 180.5)
        plt.xlabel("angle (degrees)")
    elif name == "distance":
        plt.xlim(0, 2.05)
        plt.xlabel("distance")
    plt.title(
        "Pairs of points in sphere in $\mathbb{R}^{"
        + str(n)
        + "}$"
        + " with a certain {} between them".format(name)
    )
    plt.ylabel("count")


def distance_histogram_factory(n, m, sample_in_sphere, distance_from_origin):
    """
    n: number of dimensions
    m: number of points to sample
    """
    distances = []

    ## Use a previously defined function to generate
    ## m sample points from R^n
    points = sample_in_sphere(n, m)  # SOLUTION

    for i in range(points.shape[1]):

        ## Obtain a n-length one-dimensional ndarray that represents the ith point sampled from R^n
        point = points[:, i]  # SOLUTION

        ## Calculate the distance between point and the origin
        distance = distance_from_origin(point)  # SOLUTION

        distances.append(distance)

    print("The mean of the distances in R^{} is {}".format(n, np.mean(distances)))
    print("The stdev of the distances in R^{} is {}".format(n, np.std(distances)))
    plt.hist(distances)
    plt.xlim(0, 1.05)
    plt.ylabel("count")
    plt.xlabel("distance")
    plt.title(
        "Points in sphere in $\mathbb{R}^{"
        + str(n)
        + "}$ with a certain distance from origin"
    )



import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np


def init_updating_plot(title, xlabel, ylabel, xlim, ylim):
    output_widget = widgets.Output()
    figure = plt.figure()
    axis = figure.add_subplot(111)
    axis.set_title(title)
    axis.set_xlabel(xlabel)
    axis.set_ylabel(ylabel)
    axis.set_xlim(xlim)
    axis.set_ylim(ylim)
    return output_widget, figure, axis


from ..utils import ball_to_box_ratio


def updating_ball_to_box_ratio_plot(
    callback, output_widget, figure, axis, trials=10000
):
    """
    callback is the dart_did_land_in_ball function students define in the notebook
    """
    xs = np.arange(1, 101)
    (line,) = axis.plot(0, 0)
    ratios = []

    with output_widget:
        for x in range(1, 101):
            output_widget.clear_output(wait=True)
            ratio = ball_to_box_ratio(trials, lambda: callback(x))
            ratios.append(ratio)
            line.set_xdata(xs[:x])
            line.set_ydata(ratios)
            figure.canvas.draw()



from .linear_functions import addition_widget
from .linear_functions import scalar_multiplication_widget
from .norms import p_ball_widget
from .norms import p_ball_widget_3d
from .updating_plots import init_updating_plot
from .updating_plots import updating_ball_to_box_ratio_plot



import numpy as np
import pandas as pd


def typecheck_array(array, name="array"):
    assert isinstance(
        array, np.ndarray
    ), "{} must be represented as a NumPy array".format(name)


def typecheck_vector(vector, name="vector"):
    typecheck_array(vector, name=name)
    assert (
        len(vector.shape) == 1
    ), "{} must be represented as a one-modal NumPy array".format(name)


def typecheck_unit_vector(vector, name="vector"):
    typecheck_vector(vector, name=name)
    assert np.isclose(np.linalg.norm(vector), 1), "{} must have unit norm".format(name)


def typecheck_vector_pair(
    vector1, vector2, name1="first vector", name2="second vector"
):
    typecheck_vector(vector1)
    typecheck_vector(vector2)
    assert len(vector1) == len(vector2), "{} and {} must have the same length".format(
        name1, name2
    )


def typecheck_matrix(matrix, name="matrix"):
    typecheck_array(matrix, name=name)
    assert (
        len(matrix.shape) == 2
    ), "{} must be represented as a two-modal NumPy array".format(name)


def typecheck_column_vector(column, name="column"):
    typecheck_matrix(column, name=name)
    assert (column.shape)[
        1
    ] == 1, "{} must be represented as a column vector, i.e. a matrix with a single column".format(
        name
    )


def typecheck_row_vector(row, name="row"):
    typecheck_matrix(row, name=name)
    assert (row.shape)[
        0
    ] == 1, "{} must be represented as a row vector, i.e. a matrix with a single row".format(
        name
    )


def typecheck_square_matrix(matrix, name="matrix"):
    typecheck_matrix(matrix, name=name)
    assert (matrix.shape)[0] == (matrix.shape)[
        1
    ], "In order for {} to be square, its height and width must be the same".format(
        name
    )


def typecheck_symmetric_matrix(matrix, name="matrix"):
    typecheck_square_matrix(matrix, name=name)
    assert np.allclose(
        matrix, matrix.T
    ), "In order for {} to be symmetric, it must equal its own transpose".format(name)


def typecheck_basis(basis, dimension=None):
    """
    `basis` should be list or NumPy array of vectors, i.e. one-dimensional NumPy arrays
    No check for linear independence because of laziness and not wanting to waste compute.
    """
    if dimension:
        assert len(basis) == dimension
    else:
        dimension = len(basis)
    assert isinstance(basis, (list, np.ndarray))
    for vector in basis:
        typecheck_vector(vector)
        assert len(vector) == dimension


def typecheck_dataframe(dataset, name="dataset"):
    assert isinstance(
        dataset, pd.DataFrame
    ), "{} must be represented as a Pandas DataFrame".format(name)


def typecheck_dataframe_row(dataset_row, name="dataset row"):
    typecheck_dataframe(dataset_row, name=name)
    assert (
        len(row.index) == 1
    ), "{} must be represented by a Pandas DataFrame with a single row".format(name)


def typecheck_series(dataset, name="dataset"):
    assert isinstance(
        dataset, pd.Series
    ), "{} must be represented by a Pandas Series".format(name)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..tests import typecheck_basis
from ..tests import typecheck_matrix
from .vectors import plot_vector


def plot_region_with_func_random_pts(linear_function, n=10000, plot_size=5):
    typecheck_matrix(linear_function)
    pts = np.random.rand(2, n) * 2 - 1
    fig, axs = plt.subplots(1, 2, subplot_kw={"aspect": "equal"})

    xs = pts[0, :]
    ys = pts[1, :]
    axs[0].set_xlim(-plot_size, plot_size)
    axs[0].set_ylim(-plot_size, plot_size)
    axs[0].scatter(xs, ys)
    axs[0].set_xlabel("$x_1$ value")
    axs[0].set_ylabel("$x_2$ value")
    axs[0].set_title("Before linear function")

    img = linear_function @ pts
    x_img = img[0, :]
    y_img = img[1, :]
    axs[1].set_xlim(-plot_size, plot_size)
    axs[1].set_ylim(-plot_size, plot_size)
    axs[1].scatter(x_img, y_img)
    axs[1].set_xlabel("$x_1$ value")
    axs[1].set_ylabel("$x_2$ value")
    axs[1].set_title("After linear function")

    # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.subplots_adjust.html
    plt.subplots_adjust(wspace=0.5)


def plot_circular_region_with_func_random_pts(function, n=5000):
    random_pts = pd.DataFrame(2 * np.random.rand(2, n) - 1)
    unit_circle_mask = list(random_pts.apply(np.linalg.norm) < 1)
    random_circle_pts = random_pts.iloc[:, unit_circle_mask]
    random_func_pts = random_circle_pts.apply(function)
    fig, ax = plt.subplots(1, 1, subplot_kw={"aspect": "equal"})
    ax.axhline(linewidth=1, color="black")
    ax.axvline(linewidth=1, color="black")
    ax.scatter(random_func_pts.iloc[0, :], random_func_pts.iloc[1, :])
    ax.set_xlabel("$x_1$ value")
    ax.set_ylabel("$x_2$ value")
    ax.set_title("Effect on the unit disc")


def plot_region_with_func(
    linear_function,
    title="Vectors after being modified by linear function",
    colors=["r", "y", "g", "b"],
    basis_scale=1.5,
    alpha=0.3,
    plot_axis=True,
    fig=None,
    ax=None,
    plot_size=5,
):

    if not fig or not ax:
        fig, ax = plt.subplots(subplot_kw={"aspect": "equal"})
    ax.set_title(title)

    typecheck_matrix(linear_function)

    quad1_x = np.array([0, 1, 1, 0])
    quad1_y = np.array([0, 0, 1, 1])
    quad2_x = np.array([0, -1, -1, 0])
    quad2_y = np.array([0, 0, 1, 1])
    quad3_x = np.array([0, -1, -1, 0])
    quad3_y = np.array([0, 0, -1, -1])
    quad4_x = np.array([0, 1, 1, 0])
    quad4_y = np.array([0, 0, -1, -1])

    xs = (quad1_x, quad2_x, quad3_x, quad4_x)
    ys = (quad1_y, quad2_y, quad3_y, quad4_y)

    ax.set_xlabel("$x_1$ value")
    ax.set_ylabel("$x_2$ value")

    X = np.array([np.hstack(xs), np.hstack(ys)])
    X = linear_function @ X
    ax.set_xlim(-plot_size, plot_size)
    ax.set_ylim(-plot_size, plot_size)
    polygons = []
    for i, x_list in enumerate(xs):
        n = len(x_list)
        region = X[:, :n]
        X = X[:, n:]
        x = region[0, :]
        y = region[1, :]
        color = colors[i % len(colors)]
        [polygon] = ax.fill(x, y, color)
        polygons.append(polygon)
    _ = [polygon.set_alpha(alpha) for polygon in polygons]

    if plot_axis:
        plot_vector(ax, linear_function[:, 0], label="e1")
        plot_vector(ax, linear_function[:, 1], label="e2")


def plot_region_with_two_basises(
    linear_function,
    title,
    alternative_basis,
    standard_basis,
    fig=None,
    ax=None,
    plot_size=5,
):
    typecheck_matrix(linear_function)
    typecheck_basis(alternative_basis, dimension=2)
    typecheck_basis(standard_basis, dimension=2)

    if not fig or not ax:
        fig, ax = plt.subplots(subplot_kw={"aspect": "equal"})

    plot_region_with_func(
        linear_function, title, plot_axis=False, fig=fig, ax=ax, plot_size=plot_size
    )

    plot_vector(ax, alternative_basis[0], color="m", label="v1")
    plot_vector(ax, alternative_basis[1], color="m", label="v2")

    plot_vector(ax, standard_basis[0], label="e1")
    plot_vector(ax, standard_basis[1], label="e2")



import os.path

import pandas as pd

HERE = os.path.dirname(__file__)

circuit_courts = pd.read_csv(os.path.join(HERE, "circuit_courts.csv"))

discretized_functions = pd.read_csv(os.path.join(HERE, "discretized_functions.csv"))

eastern_cities = pd.read_csv(os.path.join(HERE, "eastern_cities.csv"))

iris = pd.read_csv(os.path.join(HERE, "iris.csv"))

federal_reserve_districts = pd.read_csv(
    os.path.join(HERE, "federal_reserve_districts.csv")
)

metropolises = pd.read_csv(os.path.join(HERE, "metropolises.csv"))

mountains = pd.read_csv(os.path.join(HERE, "mountains.csv"))

states = pd.read_csv(os.path.join(HERE, "states.csv"))

students = pd.read_csv(os.path.join(HERE, "students.csv"))

western_cities = pd.read_csv(os.path.join(HERE, "western_cities.csv"))



from .histograms import distance_histogram_factory
from .histograms import pairs_histogram_factory
from .monte_carlo import p_ball_plot
from .pca_2D import cumulative_variance_plot_factory
from .pca_2D import scree_plot_factory
from .pca_3D import plot_3D_with_arrows
from .plotly import p_ball_plot_3d
from .plotly import plot_vectors_in_space
from .plotly import scatter_3D
from .plotly import scatter_3D_collinear
from .region import plot_circular_region_with_func_random_pts
from .region import plot_region_with_func
from .region import plot_region_with_func_random_pts
from .region import plot_region_with_two_basises
from .spectral_3D import plot_3D_spectral_demo
from .two_dimensional import abline
from .two_dimensional import timeseries_plot
from .vectors import does_function_preserve_addition
from .vectors import does_function_preserve_scalar_multiplication
from .vectors import plot_vector



import pathlib

from setuptools import find_namespace_packages
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="linalg_for_datasci",
    version="1.2.2",
    description="Code supporting the computational instruction for the course STAT 89A: Linear Algebra for Data Science at UC Berkeley",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/stat-89a/linalg_for_datasci",
    author="William Krinsman",
    author_email="krinsman@berkeley.edu",
    license="BSD",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
    ],
    packages=find_namespace_packages(),
    include_package_data=True,
    install_requires=["ipywidgets", "matplotlib", "numpy", "pandas", "plotly",],
)



import numpy as np

from ..arrays import col_to_vec
from ..arrays import vec_to_col
from ..tests import typecheck_basis
from ..tests import typecheck_symmetric_matrix


def Rx(theta):
    return np.array(
        [
            [1, 0, 0],
            [0, np.cos(theta), -np.sin(theta)],
            [0, np.sin(theta), np.cos(theta)],
        ]
    )


def Ry(theta):
    return np.array(
        [
            [np.cos(theta), 0, np.sin(theta)],
            [0, 1, 0],
            [-np.sin(theta), 0, np.cos(theta)],
        ]
    )


def Rz(theta):
    return np.array(
        [
            [np.cos(theta), -np.sin(theta), 0],
            [np.sin(theta), np.cos(theta), 0],
            [0, 0, 1],
        ]
    )


def get_2D_rotation_matrix(deg):
    theta = np.radians(deg)
    c, s = np.cos(theta), np.sin(theta)
    R = np.array([[c, -s], [s, c]])
    return R


# 2D spectral demo


def angle_of(vector):
    return np.arctan2(vector[1], vector[0])


def get_angle_of_rot_mat(U):
    return np.degrees(angle_of(U[:, 0]))


def get_parts_for_diagonalization_demo(A):
    typecheck_symmetric_matrix(A)

    d, V = np.linalg.eigh(A)
    d = np.flip(d, axis=0)
    V = np.flip(V, axis=1)
    V *= np.sign(V[0, 0])
    angle = get_angle_of_rot_mat(V)

    v1 = V[:, 0]
    v2 = V[:, 1]
    e1 = np.array([1, 0])
    e2 = np.array([0, 1])

    return d, V, v1, v2, e1, e2, angle


def apply_func_to_basis(linear_function, basis):
    typecheck_basis(basis)
    transformed_basis = []
    for vector in basis:
        transformed_vector = linear_function @ vec_to_col(vector)
        transformed_basis.append(col_to_vec(transformed_vector))
    typecheck_basis(transformed_basis)
    return transformed_basis



import matplotlib.pyplot as plt
import numpy as np

from ..tests import typecheck_vector


def plot_vector(ax, vector, scale=1, color="k", label="", origin=np.array([0, 0])):
    typecheck_vector(origin)
    assert len(origin) == 2
    typecheck_vector(vector)
    assert len(vector) == 2

    shifted_and_scaled_vector = (vector + origin) * scale
    scaled_vector = vector * scale

    if label:
        halign = "left" if vector[0] >= 0 else "right"
        valign = "bottom" if vector[1] >= 0 else "top"
        text_options = {
            "horizontalalignment": halign,
            "verticalalignment": valign,
            "color": color,
            "size": 22,
        }
        ax.text(*shifted_and_scaled_vector, label, **text_options)

    arrow_options = {
        "width": 0.3 * scale,
        "head_width": 0.4 * scale,
        "head_length": 0.3 * scale,
        "length_includes_head": True,
        "color": color,
    }
    return ax.arrow(*origin, *scaled_vector, **arrow_options)


def does_function_preserve_scalar_multiplication(
    angle, scalar, function, vector_name="$v$", scalar_name="$s$", function_name="$f$"
):
    fig, plots = plt.subplots(ncols=5, subplot_kw={"aspect": "equal"}, figsize=(50, 10))

    vector = np.array([np.cos(angle), np.sin(angle)])
    v, s, f = (vector, scalar, function)
    plot_size = (
        np.max(
            np.abs(
                [
                    v[0],
                    v[1],
                    f(v)[0],
                    f(v)[1],
                    (s * v)[0],
                    (s * v)[1],
                    f(s * v)[0],
                    f(s * v)[1],
                    (s * f(v))[0],
                    (s * f(v))[1],
                ]
            )
        )
        + 2
    )

    for i in range(0, 5):
        plots[i].set_xlim(-plot_size, plot_size)
        plots[i].set_ylim(-plot_size, plot_size)
        plots[i].set_xlabel("$x_1$ value")
        plots[i].set_ylabel("$x_2$ value")
        plots[i].axhline(linewidth=1, color="black")
        plots[i].axvline(linewidth=1, color="black")

    # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.subplots_adjust.html
    plt.subplots_adjust(wspace=0.3)

    plot_vector(plots[0], vector, color="k", label=vector_name)
    plots[0].set_title("Individual vector")

    plot_vector(
        plots[1],
        scalar * vector,
        color="k",
        label="{}$\cdot${}".format(scalar_name, vector_name),
    )
    plots[1].set_title("Scaled vector")

    plot_vector(
        plots[2],
        function(vector),
        color="m",
        label="{}({})".format(function_name, vector_name),
    )
    plots[2].set_title("{} applied to individual vector".format(function_name))

    plot_vector(
        plots[3],
        scalar * function(vector),
        color="m",
        label="{}$\cdot${}({}))".format(scalar_name, function_name, vector_name),
    )
    plots[3].set_title("Function first, scaling second")

    plot_vector(
        plots[4],
        function(scalar * vector),
        color="m",
        label="{}({}$\cdot${})".format(function_name, scalar_name, vector_name),
    )
    plots[4].set_title("Scaling first, function second")


def does_function_preserve_addition(
    angle1,
    angle2,
    function,
    vector1_name="$v_1$",
    vector2_name="$v_2$",
    function_name="$f$",
    head_to_tail=False,
):
    fig, plots = plt.subplots(ncols=5, subplot_kw={"aspect": "equal"}, figsize=(50, 10))

    vector1 = np.array([np.cos(angle1), np.sin(angle1)])
    vector2 = np.array([np.cos(angle2), np.sin(angle2)])

    v1, v2, f = (vector1, vector2, function)
    plot_size = (
        np.max(
            np.abs(
                [
                    v1[0],
                    v1[1],
                    f(v1)[0],
                    f(v1)[1],
                    v2[0],
                    v2[1],
                    f(v2)[0],
                    f(v2)[1],
                    (v1 + v2)[0],
                    (v1 + v2)[1],
                    f(v1 + v2)[0],
                    f(v1 + v2)[1],
                    (f(v1) + f(v2))[0],
                    (f(v1) + f(v2))[1],
                ]
            )
        )
        + 2
    )

    for i in range(0, 5):
        plots[i].set_xlim(-plot_size, plot_size)
        plots[i].set_ylim(-plot_size, plot_size)
        plots[i].set_xlabel("$x_1$ value")
        plots[i].set_ylabel("$x_2$ value")
        plots[i].axhline(linewidth=1, color="black")
        plots[i].axvline(linewidth=1, color="black")

    # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.subplots_adjust.html
    plt.subplots_adjust(wspace=0.3)

    plot_vector(plots[0], vector1, color="black", label=vector1_name)
    origin = vector1 if head_to_tail else np.array([0, 0])
    plot_vector(plots[0], vector2, color="black", label=vector2_name, origin=origin)
    plots[0].set_title("Individual vectors")

    plot_vector(
        plots[1],
        vector1 + vector2,
        color="black",
        label="{}+{}".format(vector1_name, vector2_name),
    )
    plots[1].set_title("Sum of vectors")

    plot_vector(
        plots[2],
        function(vector1),
        color="m",
        label="{}({})".format(function_name, vector1_name),
    )
    origin = function(vector1) if head_to_tail else np.array([0, 0])
    plot_vector(
        plots[2],
        function(vector2),
        color="m",
        label="{}({})".format(function_name, vector2_name),
        origin=origin,
    )
    plots[2].set_title("{} applied to individual vectors".format(function_name))

    plot_vector(
        plots[3],
        function(vector1) + function(vector2),
        color="m",
        label="{}({})+{}({})".format(
            function_name, vector1_name, function_name, vector2_name
        ),
    )
    plots[3].set_title("Function first, sum second")

    plot_vector(
        plots[4],
        function(vector1 + vector2),
        color="m",
        label="{}({}+{})".format(function_name, vector1_name, vector2_name),
    )
    plots[4].set_title("Sum first, function second")
